/**
 *  joola.io
 *
 *  Copyright Joola Smart Solutions, Ltd. <info@joo.la>
 *
 *  Licensed under GNU General Public License 3.0 or later.
 *  Some rights reserved. See LICENSE, AUTHORS.
 *
 *  @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
 */

var
  router = require('./index'),
  url = require('url'),

  _metrics = require('../lib/objects/metrics'),
  _datatables = require('../lib/objects/datatables'),
  _datasources = require('../lib/objects/datasources'),
  auth = require('../lib/auth/manager'),
  connector = require('../lib/connectors/connector');

/**
 * Lists all available dimensions. You can provide parameters and ask for a specific data source and datatable.
 *
 * The returned list will contain only dimensions authorized for the authenticated user.
 *
 * @method list
 * @param {string} [datasourceid] Datasource ID.
 * @param {string} [datatableid] Datatable ID.
 * @return {object} json structure listing all dimensions.
 */
exports.index = exports.list = {
  name: 'metrics/list',
  description: 'i list all metrics available',
  inputs: {
    required: [],
    optional: ['datasourceid', 'datatableid']
  },
  outputExample: {},
  permission: ['access_system'],
  run: function (req, res) {
    var response = {};

    if (!req.user)
      return router.responseError(new router.AuthErrorTemplate('Missing user token'), req, res);

    var metrics = _metrics.list(req.params.datasourceid, req.params.datatableid);

    if (!metrics)
      return router.responseError(new router.ErrorTemplate('Failed to find metric'), req, res);

    var __metrics = [];
    _.each(metrics, function (m) {
      if (auth.hasRole(m.roles, req.user._roles)) {
        __metrics.push(m);
      }
    });

    response.metrics = __metrics;
    return router.responseSuccess(response, req, res);
  }
};

/**
 * Lists all available dimensions. You can provide parameters and ask for a specific data source and datatable.
 *
 * The returned list will contain only dimensions authorized for the authenticated user.
 *
 * @method get
 * @param {string} [dimensionid] Dimension ID.
 * @return {object} json structure holding the dimension details.
 */
exports.get = {
  name: 'metrics/get',
  description: 'i get a specific metric',
  inputs: {
    required: ['id'],
    optional: []
  },
  outputExample: {},
  permission: ['access_system'],
  run: function (req, res) {
    var response = {};
    response.metric = null;

    if (!req.user)
      return router.responseError(new router.AuthErrorTemplate('Missing user token'), req, res);

    var metric = _metrics.get(req.params.id);
    if (!metric)
      return router.responseError(new router.ErrorTemplate('Failed to find metric'), req, res);

    if (auth.hasRole(metric.roles, req.user._roles))
      response.metric = metric;
    return router.responseSuccess(response, req, res);
  }
};

/**
 * Validates the dimension database integrity.
 *
 * @method get
 * @param {string} [dimensionid] Dimension ID.
 * @return {object} json structure holding the validation details.
 */
exports.validate = {
  name: 'metrics/validate',
  description: 'i validate a specific metric',
  inputs: {
    required: ['id'],
    optional: []
  },
  outputExample: {},
  permission: ['manage_system'],
  run: function (req, res) {
    var response = {};
    var metric = _metrics.get(req.params.id);
    var datatable = _datatables.get(metric.datatableid);
    var datasource = _datasources.get(datatable.datasourceid);
    var query = connector.createQuery();

    response.metric = ce.clone(metric);
    response.validated = false;

    query.sql = 'SELECT ' + metric.column + ' FROM ' + datatable.name + ';';
    query.limit = 1;
    query.datasource = datasource;
    connector.executeQuery(query, function (query, rows, fields, error) {
      response.queryInformation = ce.clone(query);
      if (error)
        throw error;
      else {
        response.validated = true;
      }
      return router.responseSuccess(response, req, res);
    });
  }
};

exports.values = {
  name: 'metrics/values',
  description: 'i query for all values available for a metric (limit 100)',
  inputs: {
    required: ['metricid'],
    optional: []
  },
  outputExample: {},
  permission: ['manage_system'],
  run: function (req, res) {
    var response = {};

    var metric = _metrics.get(req.params.metricid);
    var datatable = _datatables.get(metric.datatableid);
    var datasource = _datasources.get(datatable.datasourceid);
    var query = connector.createQuery();

    response.metric = ce.clone(metric);

    query.sql = 'SELECT ' + metric.column + ' FROM ' + datatable.name + ';';
    query.limit = 100;
    query.datasource = datasource;
    connector.executeQuery(query, function (query, rows, fields, error) {
      response.queryInformation = ce.clone(query);
      if (error)
        throw error;
      else {
        response.values = [];
        if (typeof(rows) == 'Array') {
          rows.forEach(function (row) {
            response.values.push(row[metric.column]);
          });
        }
      }
      return router.responseSuccess(response, req, res);
    });
  }
};